#ifndef PREVIEWCONTROLLER_H
#define PREVIEWCONTROLLER_H

#include <memory>
#include <vector>
#include "../PhotomatixMisc/photomatixconstants.h"
#include <sigc++/connection.h>
#include <gtkmm.h>
#include "presetthumbnail.h"


namespace PhotomatixImaging
{
	class ToneMapper;
	class ExposureFusionProcessor;
	class PhotoImage;
	class HdrGenerator;
	class PhotoProcessor;
	class PhotoImage;
	class XmpSettings;
}

namespace Photomatix
{
	struct Size
	{
		int width;
		int height;
	};
	
	/// Forward declarations
	class SettingsWindow;
	class PresetsWindow;
	class PreviewWindow;
	class ProgressForm;
	
	
	/** Controller responsible for preview mode. It creates all preview mode windows (Settings, Presets and Preview),
	 * is responsible for preview updates as well as final tone-mapping/fusion.
	 */
	class PreviewController
	{
	private:
		volatile bool isPreviewUpdating;
		
		int zoomInLevels;
		int zoomOutLevels;
		int currentZoomLevel;
		
		int currentMethod;
		
		int basePreviewContrastOptimizerW;
		int basePreviewContrastOptimizerH;
		int basePreviewW;
		int basePreviewH;
		
		bool builtInPresetsSelected;
		
		float currentSettings[PhotomatixMisc::XmpParamsLength];
		
		/// Array of preview sizes for all zoom levels. Elements of this array are calculated based on the input image dimensions
		std::vector <Size> previewSizes;
		
		/// Tone-mapper and blenders used by the application
		std::shared_ptr <PhotomatixImaging::ToneMapper> mapper;
		std::shared_ptr <PhotomatixImaging::ExposureFusionProcessor> blender;
		
		
		PhotomatixImaging::PhotoProcessor* processor;
		
		/// Image buffer used to create preview and thumbnails for exposure fusion
		std::shared_ptr <PhotomatixImaging::HdrGenerator> hdrGenerator;
		
		PhotomatixImaging::XmpSettings previousContrastOptimizerSettings;
		
		std::shared_ptr <PresetsWindow> presetsForm;
		std::shared_ptr <SettingsWindow> settingsForm;
		PreviewWindow* previewForm;
		ProgressForm* progressForm;
		
		sigc::connection mapperDone;
		sigc::connection blenderDone;
		sigc::connection finalProcessorDone;
		sigc::connection finalProcessorCancelled;
		Glib::Mutex sharedObject;
		
		std::shared_ptr <PhotomatixImaging::PhotoImage> originalPhoto;
		/// PhotoImage used as a source image for tone-mapping - typically it's a downsampled version of a original photo.
		std::vector <std::shared_ptr <PhotomatixImaging::PhotoImage> > basePreviewPhotos;
		/// Image buffers used as a source for exposure fusion
		std::vector <std::shared_ptr <std::vector<unsigned char>> > basePreviewBuffers;
		std::vector <std::shared_ptr <std::vector<unsigned char>> > ghostMasks;
		std::vector <std::shared_ptr <PhotomatixImaging::PhotoImage>> fusionThumbImages;
		
		std::shared_ptr <PhotomatixImaging::PhotoImage> basePreviewInputContrastOptimizer;
		
		
		std::shared_ptr<std::vector<unsigned char>> imageBuffer;
		std::shared_ptr <PhotomatixImaging::PhotoImage> thumbnailSourcePhoto;
		std::shared_ptr <std::vector <unsigned char> > ghostMask;
		std::shared_ptr <std::vector <unsigned char>> thumbnailImageBuffer;
		
	protected:
		void OnPreviewToneMappingDone();
		void OnPreviewFusingDone();
		void OnSettingsChanged(PhotomatixImaging::XmpSettings* settings);
		void OnThumbnailChanged(PhotomatixImaging::Preset* preset);
		void OnPreviewZoomLevelChanged(int zoomLevel);
		void OnPresetChanged(int preset, std::string fileName);
		void OnPreviewClosed();
		void OnFinalProcessing();
		void OnFinalProcessingDone();
		void OnFinalProcessingCancelled();
		void OnPresetsCategorySelected(bool builtIn);
		void OnThumbnailsRenderingChanged(bool done);
		
		
		void GetPresetThumbnailsSize(int inWidth, int inHeight, int& outWidth, int& outHeight);
		void InitThumbnails();
		
		void InitPreviewSizes();
		void InitPreviewImage();
		
		void GetBasePreviewSizeForContrastOptimizer(int wPreview, int hPreview, int wOrig, int hOrig, int& basePreviewW, int& basePreviewH);
	
	public:
		PreviewController(std::shared_ptr <PhotomatixImaging::PhotoImage> photo, std::shared_ptr <PhotomatixImaging::HdrGenerator> generator);
		~PreviewController();
		
		void SettingsChanged() {}
		void UpdatePreviewImage();
		void UpdatePreviewImageFinished() {}
		void UpdatePreviewImageStopped() {}
		void FinalProcessing() {}
		
		void SetPhotoImage(std::shared_ptr <PhotomatixImaging::PhotoImage> photo) {this->originalPhoto = photo;}
		
		int GetCurrentMethod() const {return this->currentMethod;}
		float* GetCurrentSettings() {return this->currentSettings;}
		
		std::shared_ptr <PhotomatixImaging::HdrGenerator> GetHdrGenerator() {return this->hdrGenerator;}
		std::shared_ptr <PhotomatixImaging::PhotoImage> GetOriginalPhoto() {return this->originalPhoto;}
		
		sigc::signal <void> signal_preview_closed;
		sigc::signal <void, std::shared_ptr<PhotomatixImaging::PhotoImage> > signal_preview_finished;
	};
}

#endif // PREVIEWCONTROLLER_H
